@@ -1,24 +1,37 @@
 #!/usr/bin/env python
 
 """Humanizing functions for numbers."""
+from __future__ import annotations
 
 import math
 import re
+import sys
 from fractions import Fraction
+from typing import TYPE_CHECKING
 
 from .i18n import _gettext as _
 from .i18n import _ngettext
 from .i18n import _ngettext_noop as NS_
 from .i18n import _pgettext as P_
-from .i18n import thousands_separator
+from .i18n import decimal_separator, thousands_separator
+
+if TYPE_CHECKING:
+    if sys.version_info >= (3, 10):
+        from typing import TypeAlias
+    else:
+        from typing_extensions import TypeAlias
+
+# This type can be better defined by typing.SupportsInt, typing.SupportsFloat
+# but that's a Python 3.8 only typing option.
+NumberOrString: TypeAlias = "float | str"
 
 
-def ordinal(value, gender="male"):
+def ordinal(value: NumberOrString, gender: str = "male") -> str:
     """Converts an integer to its ordinal as a string.
 
     For example, 1 is "1st", 2 is "2nd", 3 is "3rd", etc. Works for any integer or
-    anything `int()` will turn into an integer. Anything other value will have nothing
-    done to it.
+    anything `int()` will turn into an integer. Anything else will return the output
+    of str(value).
 
     Examples:
         ```pycon
@@ -38,7 +51,7 @@ def ordinal(value, gender="male"):
         '111th'
         >>> ordinal("something else")
         'something else'
-        >>> ordinal(None) is None
+        >>> ordinal([1, 2, 3]) == "[1, 2, 3]"
         True
 
         ```
@@ -52,7 +65,7 @@ def ordinal(value, gender="male"):
     try:
         value = int(value)
     except (TypeError, ValueError):
-        return value
+        return str(value)
     if gender == "male":
         t = (
             P_("0 (male)", "th"),
@@ -84,7 +97,7 @@ def ordinal(value, gender="male"):
     return f"{value}{t[value % 10]}"
 
 
-def intcomma(value, ndigits=None):
+def intcomma(value: NumberOrString, ndigits: int | None = None) -> str:
     """Converts an integer to a string containing commas every three digits.
 
     For example, 3000 becomes "3,000" and 45000 becomes "45,000". To maintain some
@@ -104,8 +117,10 @@ def intcomma(value, ndigits=None):
         '1,234.55'
         >>> intcomma(14308.40, 1)
         '14,308.4'
-        >>> intcomma(None) is None
-        True
+        >>> intcomma("14308.40", 1)
+        '14,308.4'
+        >>> intcomma(None)
+        'None'
 
         ```
     Args:
@@ -113,27 +128,32 @@ def intcomma(value, ndigits=None):
         ndigits (int, None): Digits of precision for rounding after the decimal point.
 
     Returns:
-        str: string containing commas every three digits.
+        str: String containing commas every three digits.
     """
-    sep = thousands_separator()
+    thousands_sep = thousands_separator()
+    decimal_sep = decimal_separator()
     try:
         if isinstance(value, str):
-            float(value.replace(sep, ""))
+            value = value.replace(thousands_sep, "").replace(decimal_sep, ".")
+            if "." in value:
+                value = float(value)
+            else:
+                value = int(value)
         else:
             float(value)
     except (TypeError, ValueError):
-        return value
+        return str(value)
 
-    if ndigits:
+    if ndigits is not None:
         orig = "{0:.{1}f}".format(value, ndigits)
     else:
         orig = str(value)
-
-    new = re.sub(r"^(-?\d+)(\d{3})", rf"\g<1>{sep}\g<2>", orig)
-    if orig == new:
-        return new
-    else:
-        return intcomma(new)
+    orig = orig.replace(".", decimal_sep)
+    while True:
+        new = re.sub(r"^(-?\d+)(\d{3})", rf"\g<1>{thousands_sep}\g<2>", orig)
+        if orig == new:
+            return new
+        orig = new
 
 
 powers = [10**x for x in (3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33, 100)]
@@ -153,7 +173,7 @@ human_powers = (
 )
 
 
-def intword(value, format="%.1f"):
+def intword(value: NumberOrString, format: str = "%.1f") -> str:
     """Converts a large integer to a friendly text representation.
 
     Works best for numbers over 1 million. For example, 1_000_000 becomes "1.0 million",
@@ -172,8 +192,8 @@ def intword(value, format="%.1f"):
         '1.2 billion'
         >>> intword(8100000000000000000000000000000000)
         '8.1 decillion'
-        >>> intword(None) is None
-        True
+        >>> intword(None)
+        'None'
         >>> intword("1234000", "%0.3f")
         '1.234 million'
 
@@ -190,28 +210,40 @@ def intword(value, format="%.1f"):
     try:
         value = int(value)
     except (TypeError, ValueError):
-        return value
+        return str(value)
+
+    if value < 0:
+        value *= -1
+        negative_prefix = "-"
+    else:
+        negative_prefix = ""
 
     if value < powers[0]:
-        return str(value)
-    for ordinal, power in enumerate(powers[1:], 1):
+        return negative_prefix + str(value)
+
+    for ordinal_, power in enumerate(powers[1:], 1):
         if value < power:
-            chopped = value / float(powers[ordinal - 1])
+            chopped = value / float(powers[ordinal_ - 1])
             if float(format % chopped) == float(10**3):
-                chopped = value / float(powers[ordinal])
-                singular, plural = human_powers[ordinal]
+                chopped = value / float(powers[ordinal_])
+                singular, plural = human_powers[ordinal_]
                 return (
-                    " ".join([format, _ngettext(singular, plural, math.ceil(chopped))])
+                    negative_prefix
+                    + " ".join(
+                        [format, _ngettext(singular, plural, math.ceil(chopped))]
+                    )
                 ) % chopped
-            else:
-                singular, plural = human_powers[ordinal - 1]
-                return (
-                    " ".join([format, _ngettext(singular, plural, math.ceil(chopped))])
-                ) % chopped
-    return str(value)
+
+            singular, plural = human_powers[ordinal_ - 1]
+            return (
+                negative_prefix
+                + " ".join([format, _ngettext(singular, plural, math.ceil(chopped))])
+            ) % chopped
+
+    return negative_prefix + str(value)
 
 
-def apnumber(value):
+def apnumber(value: NumberOrString) -> str:
     """Converts an integer to Associated Press style.
 
     Examples:
@@ -226,8 +258,8 @@ def apnumber(value):
       'seven'
       >>> apnumber("foo")
       'foo'
-      >>> apnumber(None) is None
-      True
+      >>> apnumber(None)
+      'None'
 
       ```
     Args:
@@ -235,12 +267,13 @@ def apnumber(value):
 
     Returns:
         str: For numbers 0-9, the number spelled out. Otherwise, the number. This always
-        returns a string unless the value was not `int`-able, unlike the Django filter.
+        returns a string unless the value was not `int`-able, then `str(value)`
+        is returned.
     """
     try:
         value = int(value)
     except (TypeError, ValueError):
-        return value
+        return str(value)
     if not 0 <= value < 10:
         return str(value)
     return (
@@ -257,7 +290,7 @@ def apnumber(value):
     )[value]
 
 
-def fractional(value):
+def fractional(value: NumberOrString) -> str:
     """Convert to fractional number.
 
     There will be some cases where one might not want to show ugly decimal places for
@@ -271,6 +304,7 @@ def fractional(value):
     * a string representation of a fraction
     * or a whole number
     * or a mixed fraction
+    * or the str output of the value, if it could not be converted
 
     Examples:
         ```pycon
@@ -284,8 +318,8 @@ def fractional(value):
         '1'
         >>> fractional("ten")
         'ten'
-        >>> fractional(None) is None
-        True
+        >>> fractional(None)
+        'None'
 
         ```
     Args:
@@ -297,22 +331,23 @@ def fractional(value):
     try:
         number = float(value)
     except (TypeError, ValueError):
-        return value
+        return str(value)
     whole_number = int(number)
     frac = Fraction(number - whole_number).limit_denominator(1000)
-    numerator = frac._numerator
-    denominator = frac._denominator
+    numerator = frac.numerator
+    denominator = frac.denominator
     if whole_number and not numerator and denominator == 1:
         # this means that an integer was passed in
         # (or variants of that integer like 1.0000)
         return f"{whole_number:.0f}"
-    elif not whole_number:
+
+    if not whole_number:
         return f"{numerator:.0f}/{denominator:.0f}"
-    else:
-        return f"{whole_number:.0f} {numerator:.0f}/{denominator:.0f}"
+
+    return f"{whole_number:.0f} {numerator:.0f}/{denominator:.0f}"
 
 
-def scientific(value, precision=2):
+def scientific(value: NumberOrString, precision: int = 2) -> str:
     """Return number in string scientific notation z.wq x 10ⁿ.
 
     Examples:
@@ -322,7 +357,7 @@ def scientific(value, precision=2):
         >>> scientific(int(500))
         '5.00 x 10²'
         >>> scientific(-1000)
-        '1.00 x 10⁻³'
+        '-1.00 x 10³'
         >>> scientific(1000, 1)
         '1.0 x 10³'
         >>> scientific(1000, 3)
@@ -331,8 +366,8 @@ def scientific(value, precision=2):
         '9.90 x 10¹'
         >>> scientific("foo")
         'foo'
-        >>> scientific(None) is None
-        True
+        >>> scientific(None)
+        'None'
 
         ```
 
@@ -354,35 +389,19 @@ def scientific(value, precision=2):
         "7": "⁷",
         "8": "⁸",
         "9": "⁹",
-        "+": "⁺",
         "-": "⁻",
     }
-    negative = False
     try:
-        if "-" in str(value):
-            value = str(value).replace("-", "")
-            negative = True
-
-        if isinstance(value, str):
-            value = float(value)
-
-        fmt = "{:.%se}" % str(int(precision))
-        n = fmt.format(value)
-
+        value = float(value)
     except (ValueError, TypeError):
-        return value
-
+        return str(value)
+    fmt = "{:.%se}" % str(int(precision))
+    n = fmt.format(value)
     part1, part2 = n.split("e")
-    if "-0" in part2:
-        part2 = part2.replace("-0", "-")
-
-    if "+0" in part2:
-        part2 = part2.replace("+0", "")
+    # Remove redundant leading '+' or '0's (preserving the last '0' for 10⁰).
+    part2 = re.sub(r"^\+?(\-?)0*(.+)$", r"\1\2", part2)
 
     new_part2 = []
-    if negative:
-        new_part2.append(exponents["-"])
-
     for char in part2:
         new_part2.append(exponents[char])
 
@@ -391,7 +410,14 @@ def scientific(value, precision=2):
     return final_str
 
 
-def clamp(value, format="{:}", floor=None, ceil=None, floor_token="<", ceil_token=">"):
+def clamp(
+    value: float,
+    format: str = "{:}",
+    floor: float | None = None,
+    ceil: float | None = None,
+    floor_token: str = "<",
+    ceil_token: str = ">",
+) -> str:
     """Returns number with the specified format, clamped between floor and ceil.
 
     If the number is larger than ceil or smaller than floor, then the respective limit
@@ -417,17 +443,17 @@ def clamp(value, format="{:}", floor=None, ceil=None, floor_token="<", ceil_toke
     Args:
         value (int, float): Input number.
         format (str OR callable): Can either be a formatting string, or a callable
-        function than receives value and returns a string.
+            function that receives value and returns a string.
         floor (int, float): Smallest value before clamping.
         ceil (int, float): Largest value before clamping.
         floor_token (str): If value is smaller than floor, token will be prepended
-        to output.
+            to output.
         ceil_token (str): If value is larger than ceil, token will be prepended
-        to output.
+            to output.
 
     Returns:
         str: Formatted number. The output is clamped between the indicated floor and
-        ceil. If the number if larger than ceil or smaller than floor, the output will
+        ceil. If the number is larger than ceil or smaller than floor, the output will
         be prepended with a token indicating as such.
 
     """
@@ -445,10 +471,67 @@ def clamp(value, format="{:}", floor=None, ceil=None, floor_token="<", ceil_toke
 
     if isinstance(format, str):
         return token + format.format(value)
-    elif callable(format):
+
+    if callable(format):
         return token + format(value)
+
+    raise ValueError(
+        "Invalid format. Must be either a valid formatting string, or a function "
+        "that accepts value and returns a string."
+    )
+
+
+def metric(value: float, unit: str = "", precision: int = 3) -> str:
+    """Return a value with a metric SI unit-prefix appended.
+
+    Examples:
+        ```pycon
+        >>> metric(1500, "V")
+        '1.50 kV'
+        >>> metric(2e8, "W")
+        '200 MW'
+        >>> metric(220e-6, "F")
+        '220 μF'
+        >>> metric(1e-14, precision=4)
+        '10.00 f'
+
+        ```
+
+    The unit prefix is always chosen so that non-significant zero digits are required.
+    i.e. `123,000` will become `123k` instead of `0.123M` and `1,230,000` will become
+    `1.23M` instead of `1230K`. For numbers that are either too huge or too tiny to
+    represent without resorting to either leading or trailing zeroes, it falls back to
+    `scientific()`.
+    ```pycon
+    >>> metric(1e40)
+    '1.00 x 10⁴⁰'
+
+    ```
+
+    Args:
+        value (int, float): Input number.
+        unit (str): Optional base unit.
+        precision (int): The number of digits the output should contain.
+
+    Returns:
+        str:
+    """
+    exponent = int(math.floor(math.log10(abs(value)))) if value != 0 else 0
+
+    if exponent >= 27 or exponent < -24:
+        return scientific(value, precision - 1) + unit
+
+    value /= 10 ** (exponent // 3 * 3)
+    if exponent >= 3:
+        ordinal_ = "kMGTPEZY"[exponent // 3 - 1]
+    elif exponent < 0:
+        ordinal_ = "mμnpfazy"[(-exponent - 1) // 3]
     else:
-        raise ValueError(
-            "Invalid format. Must be either a valid formatting string, or a function "
-            "that accepts value and returns a string."
-        )
+        ordinal_ = ""
+    value_ = format(value, ".%if" % (precision - (exponent % 3) - 1))
+    if not (unit or ordinal_) or unit in ("°", "′", "″"):
+        space = ""
+    else:
+        space = " "
+
+    return f"{value_}{space}{ordinal_}{unit}"
